const coral = @import("coral");

const kym = @import("../kym.zig");

associative: RefTable,
contiguous: RefList,

const RefList = coral.list.Stack(?*kym.RuntimeRef);

const RefTable = coral.map.Table(*kym.RuntimeRef, *kym.RuntimeRef, struct {
	pub const hash = kym.RuntimeRef.hash;

	pub const equals = kym.RuntimeRef.equals;
});

const Self = @This();

pub fn free(self: *Self, env: *kym.RuntimeEnv) void {
	{
		var field_iterable = self.associative.as_iterable();

		while (field_iterable.next()) |entry| {
			env.discard(entry.key);
			env.discard(entry.value);
		}
	}

	self.associative.free();

	while (self.contiguous.pop()) |value| {
		if (value) |ref| {
			env.discard(ref);
		}
	}

	self.contiguous.free();
}

pub fn make(env: *kym.RuntimeEnv) Self {
	return .{
		.associative = RefTable.make(env.allocator, .{}),
		.contiguous = RefList.make(env.allocator),
	};
}

pub const typeinfo = &kym.Typeinfo{
	.size = @sizeOf(Self),
	.name = "table",
	.destruct = typeinfo_destruct,
	.get = typeinfo_get,
	.set = typeinfo_set,
};

fn typeinfo_destruct(env: *kym.RuntimeEnv, userdata: []coral.io.Byte) void {
	@as(*Self, @ptrCast(@alignCast(userdata))).free(env);
}

fn typeinfo_get(env: *kym.RuntimeEnv, userdata: []coral.io.Byte, index: *const kym.RuntimeRef) kym.RuntimeError!?*kym.RuntimeRef {
	const table = @as(*Self, @ptrCast(@alignCast(userdata)));
	const acquired_index = index.acquire();

	defer env.discard(acquired_index);

	if (acquired_index.as_fixed()) |fixed| {
		if (fixed < 0) {
			// TODO: Negative indexing.
			unreachable;
		}

		if (fixed < table.contiguous.values.len) {
			return (table.contiguous.values[@intCast(fixed)] orelse return null).acquire();
		}
	}

	if (table.associative.lookup(acquired_index)) |value| {
		return value.acquire();
	}

	return null;
}

fn typeinfo_set(env: *kym.RuntimeEnv, userdata: []coral.io.Byte, index: *const kym.RuntimeRef, value: ?*const kym.RuntimeRef) kym.RuntimeError!void {
	const table = @as(*Self, @ptrCast(@alignCast(userdata)));
	const acquired_index = index.acquire();

	errdefer env.discard(acquired_index);

	if (acquired_index.as_fixed()) |fixed| {
		if (fixed < 0) {
			// TODO: Negative indexing.
			unreachable;
		}

		if (fixed < table.contiguous.values.len) {
			const maybe_replacing = &table.contiguous.values[@intCast(fixed)];

			if (maybe_replacing.*) |replacing| {
				env.discard(replacing);
			}

			maybe_replacing.* = if (value) |ref| ref.acquire() else null;

			return;
		}
	}

	const acquired_value = (value orelse {
		if (table.associative.remove(acquired_index)) |removed| {
			env.discard(removed.key);
			env.discard(removed.value);
		}

		return;
	}).acquire();

	errdefer env.discard(acquired_value);

	if (try table.associative.replace(acquired_index, acquired_value)) |replaced| {
		env.discard(replaced.key);
		env.discard(replaced.value);
	}
}
